@@ -158,57 +158,47 @@ def despeckle_local(M, stds=2, width=2):
 
 
 def bin_dense(M, subsampling_factor=3):
-    """Sum over each block of given subsampling factor, returns a matrix whose
-    dimensions are this much as small (e.g. a 27x27 matrix binned with a
-    subsampling factor equal to 3 will return a 9x9 matrix whose each component
-    is the sum of the corresponding 3x3 block in the original matrix).
-    Remaining columns and rows are summed likewise and added to the end of the
-    new matrix.
-
-    :note: Will not work for numpy versions below 1.7
     """
+    Wraps the bin_sparse function to apply it on dense matrices. Bins are merged
+    by groups of N to produce a lower resolution matrix.
 
-    m = min(M.shape)
-    n = (m // subsampling_factor) * subsampling_factor
+    Parameters
+    ----------
+    M : numpy.array
+        2D array containing the Hi-C contact map
+    subsampling_factor : int
+        The number of bins to include in each group (subsample).
 
-    if n == 0:
-        return np.array([M.sum()])
+    Returns
+    -------
+    out_M : numpy.array
+        The subsamples matrix, with a resolution lower than the input by a defined factor.
+    """
+    S = coo_matrix(M)
+    out_S = bin_sparse(S, subsampling_factor=subsampling_factor)
+    out_M = out_S.todense()
 
-    N = np.array(M[:n, :n], dtype=np.float64)
-    N = N.reshape(
-        n // subsampling_factor,
-        subsampling_factor,
-        n // subsampling_factor,
-        subsampling_factor,
-    ).sum(axis=(1, 3))
-    if m > n:
-        remaining_row = M[n:, :n]
-        remaining_col = M[:n, n:]
-        remaining_square = M[n:m, n:m]
-        R = remaining_row.reshape(
-            m % subsampling_factor, m // subsampling_factor, subsampling_factor
-        ).sum(axis=(0, 2))
-        C = (
-            remaining_col.T.reshape(
-                m % subsampling_factor,
-                m // subsampling_factor,
-                subsampling_factor,
-            )
-            .sum(axis=(0, 2))
-            .T
-        )
-        S = remaining_square.sum()
-        N = np.append(N, [R], axis=0)
-        result = np.append(N, np.array([list(C) + [S]]).T, axis=1)
-    else:
-        result = N
-
-    return result
+    return out_M
 
 
 def bin_sparse(M, subsampling_factor=3):
-    """Perform the bin_dense procedure for sparse matrices. Remaining rows
-    and cols are put into a smaller bin at the end.
+    """
+    Bins a sparse matrix by combining bins into groups of user defined size. Binsize
+    is independent of genomic coordinates. Remaining rows and cols are put into a
+    smaller bin at the end.
+
+    Parameters
+    ----------
+    M : scipy.sparse.coo_matrix
+        The input Hi-C matrix in a sparse format.
+    subsampling_factor : int
+        The number of bins to include in each group (subsample).
+
+    Returns
+    -------
+    scipy.sparse.coo_matrix
+        The subsamples matrix, with a resolution lower than the input by a defined factor.
+        
     """
 
     N = M.tocoo()
@@ -367,7 +357,7 @@ def bin_bp_sparse(M, positions, bin_len=10000):
     chroms = np.repeat(range(len(chromlen)), chromlen)
     # Get binned positions
     positions = positions // bin_len
-    frags = np.array([chroms, positions], dtype=np.int64).T
+    frags = np.transpose(np.array([chroms, positions], dtype=np.int64))
     # Keep track of index fragments
     frag_idx = range(frags.shape[0])
     # Unique bin coordinates to create
@@ -424,11 +414,10 @@ def bin_bp_sparse(M, positions, bin_len=10000):
                 out_pos[actual_bin_No] = coords[1] * bin_len + curr_shift
         unique_bin_No += 1
         actual_bin_No += 1
-    row[np.where(r.row >= last_frag)] = actual_bin_No
-    col[np.where(r.col >= last_frag)] = actual_bin_No
+    row[np.where(r.row >= last_frag)] = actual_bin_No - 1
+    col[np.where(r.col >= last_frag)] = actual_bin_No - 1
     # Sum data of duplicate row/col pairs 
     # (i.e. combine contacts of all fragments in same bin)
-
     binned = coo_matrix((r.data, (row, col)), shape=(actual_bin_No, actual_bin_No))
     binned.sum_duplicates()
     binned.eliminate_zeros()
@@ -566,7 +555,7 @@ def get_good_bins(M, n_mad=2.0, s_min=None, s_max=None, symmetric=False):
 
     return filter_bins
 
-def trim_dense(M, n_std=3, s_min=None, s_max=None):
+def trim_dense(M, n_mad=3, s_min=None, s_max=None):
     """By default, return a matrix stripped of component
     vectors whose sparsity (i.e. total contact count on a
     single column or row) deviates more than specified number
@@ -578,7 +567,7 @@ def trim_dense(M, n_std=3, s_min=None, s_max=None):
     ----------
     M : 2D numpy array of floats
         Dense Hi-C contact matrix
-    n_std : int
+    n_mad : int
         Minimum number of standard deviation by which a the sum of
         contacts in a component vector must deviate from the mean
         to be trimmed.
@@ -596,7 +585,7 @@ def trim_dense(M, n_std=3, s_min=None, s_max=None):
     """
 
     S = coo_matrix(M)
-    S_out = trim_sparse(S, n_std=n_std, s_min=s_min, s_max=s_max)
+    S_out = trim_sparse(S, n_mad=n_mad, s_min=s_min, s_max=s_max)
     M_out = S_out.todense()
     return M_out
 
@@ -738,12 +727,18 @@ def normalize_sparse(M, norm="ICE", iterations=40, n_mad=3.0):
         Normalized sparse matrix.
     """
     # Making full symmetric matrix if not symmetric already (e.g. upper triangle)
-    r = csr_matrix(M)
-    r = r.astype(np.float64)
+    r = M.astype(np.float64)
     good_bins = get_good_bins(M, n_mad=n_mad)
     # Set values in non detectable bins to 0
-    r[~good_bins, :] = 0
-    r[:, ~good_bins] = 0
+    # For faster masking of bins, mask bins using dot product with an identity
+    # matrix where bad bins have been masked on the diagonal
+    # E.g. if removing the second bin (row and column):
+    # 1 0 0     9 6 5     1 0 0     9 0 5
+    # 0 0 0  X  6 8 7  X  0 0 0  =  0 0 0
+    # 0 0 1     6 7 8     0 0 1     6 0 8
+    mask_mat = sparse.eye(r.shape[0])
+    mask_mat.data[0][~good_bins] = 0
+    r = mask_mat.dot(r).dot(mask_mat)
     r = coo_matrix(r)
     r.eliminate_zeros()
     if norm == "ICE":
@@ -1175,7 +1170,7 @@ def to_pdb(
     if indices is None:
         indices = range(n + 1)
     if special_bins is None:
-        special_bins = np.zeros(n + 1)
+        special_bins = np.zeros(n + 1, dtype=int)
 
     structure_shapes_match = structure.shape[0] == structure.shape[1]
     if isinstance(structure, np.ndarray) and structure_shapes_match:
@@ -1192,9 +1187,9 @@ def to_pdb(
     Z = np.around(Z, 3)
 
     reference = ["OW", "OW", "CE", "TE", "tR"]
-
     with open(filename, "w") as f:
         for i in range(1, n):
+
             line = "ATOM"  # 1-4 "ATOM"
             line += "  "  # 5-6 unused
             line += str(i).rjust(5)  # 7-11 atom serial number
@@ -1308,48 +1303,55 @@ def distance_to_contact(D, alpha=1):
     return M
 
 
-def subsample_contacts(M, prop):
+def subsample_contacts(M, n_contacts):
     """Bootstrap sampling of contacts in a sparse Hi-C map.
 
     Parameters
     ----------
-    M : scipy.sparse.csr_matrix
+    M : scipy.sparse.coo_matrix
         The input Hi-C contact map in sparse format.
-    prop : float
-        The proportion of contacts to sample.
-
+    n_contacts : float
+        The number of contacts to be sampled if larger than one.
+        The proportion of contacts to be sampled if between 0 and 1.
     Returns
     -------
-    scipy.sparse.csr_matrix
+    scipy.sparse.coo_matrix
         A new matrix with a fraction of the original contacts.
     """
-    # NOTE: RAM usage vs speed could be balanced by flushing
-    # dictionary and recomputing cumsum a given number of times.
-
-    if prop > 1 or prop < 0:
-        raise ValueError(
-            "The proportion of contacts to sample must be between 0 and 1."
-        )
-    # Only work with non-zero data of matrices
-    O = csr_matrix(M).data
-    S = O.copy()
+    try:
+        if n_contacts <= 1 and n_contacts > 0:
+            n_contacts *= M.data.sum()
+        elif n_contacts < 0:
+            logger.error("n_contacts must be strictly positive")
+    except ValueError as e:
+        logger.error("n_contacts must be a float")
+        raise e
+    S = M.data.copy()
     # Match cell idx to cumulative number of contacts
-    cum_sum = np.cumsum(O)
-    # Total number of contacts to remove
-    tot_contacts = int(max(cum_sum))
-    n_remove = int((1 - prop) * tot_contacts)
-    # Store contacts that have already been removed
-    removed = {}
+    cum_counts = np.cumsum(S)
+    # Total number of contacts to sample
+    tot_contacts = int(cum_counts[-1])
+    
+    # Sample desired number of contacts from the range(0, n_contacts) array
+    sampled_contacts = np.random.choice(
+        int(tot_contacts),
+        size=int(n_contacts),
+        replace=False
+    )
+    
+    # Get indices of sampled contacts in the cum_counts array
+    idx = np.searchsorted(cum_counts, sampled_contacts, side='right')
 
-    for _ in range(n_remove):
-        to_remove = np.random.randint(tot_contacts)
-        while to_remove in removed:
-            to_remove = np.random.randint(tot_contacts)
-        # Find idx of cell to deplete and deplete it
-        removed[to_remove] = np.searchsorted(cum_sum, to_remove)
-        S[removed[to_remove]] -= 1
+    # Bin those indices to the same dimensions as matrix data to get counts
+    sampled_counts = np.bincount(idx, minlength=S.shape[0])
 
-    return csr_matrix((S, (M.row, M.col)), shape=(M.shape[0], M.shape[1]))
+    # Get nonzero values to build new sparse matrix
+    nnz_mask = sampled_counts > 0
+    sampled_counts = sampled_counts[nnz_mask].astype(np.float64)
+    sampled_rows = M.row[nnz_mask]
+    sampled_cols = M.col[nnz_mask]
+
+    return coo_matrix((sampled_counts, (sampled_rows, sampled_cols)), shape=(M.shape[0], M.shape[1]))
 
 
 def shortest_path_interpolation(matrix, alpha=1, strict=True):
